/*
 * (c) 2006-2017 Thomas Maier-Komor
 * LICENSE: GPLv2, see file COPYING for details
 */

#include <stdlib.h>
#include <strings.h>
#include <sys/types.h>
#include <unistd.h>

#include "config.h"

#ifdef HAVE_CURSES_H
#include <curses.h>
#ifdef __sun		// term.h cannot be found on solaris by autoconf
#include <term.h>	// and requires curses.h as a prerequisite
#endif
#ifndef HAVE_TERMINFO
#define HAVE_TERMINFO
#endif
#endif

#ifdef HAVE_TERM_H
#include <term.h>
#ifndef HAVE_TERMINFO
#define HAVE_TERMINFO
#endif
#endif

#ifdef HAVE_NCURSES_H
#include <ncurses.h>
#ifndef HAVE_TERMINFO
#define HAVE_TERMINFO
#endif
#endif

#ifndef ERR
#define ERR -1
#endif

#include "colortty.h"
#include "log.h"
#include "support.h"

ttymode_t TtyMode = tty_auto;

color_t
	ColorStart = cyan,
	ColorInfo = noattr,
	ColorWarn = yellow,
	ColorError = red,
	ColorFail = red,
	ColorDone = blue,
	ColorOut = noattr;

char *PfxStart = "", *PfxDone = "", *PfxFail = "", *PfxWarn = "", *PfxError = "", *PfxOut = "", *PfxInfo = "";

const char *TiSave = "", *TiRestore = "", *TiCursorLL = "", *TiClrEol = "", *TiNoAttr = "",
	*TiBlack = "", *TiBlue = "", *TiCyan = "", *TiGreen = "", *TiGrey = "",
	*TiMagenta = "", *TiRed = "", *TiWhite = "", *TiYellow = "";


static void init_tty_none()
{
	TiBlack = "";
	TiBlue = "";
	TiCyan = "";
	TiGreen = "";
	TiGrey = "";
	TiMagenta = "";
	TiRed = "";
	TiWhite = "";
	TiYellow = "";
	TiClrEol = "";
	TiNoAttr = "";
	TiCursorLL = "";
	TiSave = "";
	TiRestore = "";
}


static void init_tty_ansi()
{
	/* background: black */
	TiBlack = "\e[30m";
	TiBlue = "\e[34m";
	TiCyan = "\e[36m";
	TiGreen = "\e[32m";
	TiGrey = "\e[90m";
	TiMagenta = "\e[35m";
	TiRed = "\e[31m";
	TiWhite = "\e[37m";
	TiYellow = "\e[33m";
	TiClrEol = "\e[K";
	TiNoAttr = "\e[0m";
	TiCursorLL = "";		// TODO which is the correct escape code???
	TiSave = "\e[s";
	TiRestore = "\e[r";
}


const char *colorstr(color_t c)
{
	switch (c) {
	case black:
		return TiBlack;
	case blue:
		return TiBlue;
	case cyan:
		return TiCyan;
	case green:
		return TiGreen;
	case magenta:
		return TiMagenta;
	case red:
		return TiRed;
	case white:
		return TiWhite;
	case yellow:
		return TiYellow;
	case noattr:
		return TiNoAttr;
	default:
		abort();
	}
	return 0;
}


color_t str2color(const char *str)
{
	if (0 == strcasecmp(str,"blue"))
		return blue;
	if (0 == strcasecmp(str,"black"))
		return black;
	if (0 == strcasecmp(str,"cyan"))
		return cyan;
	if (0 == strcasecmp(str,"green"))
		return green;
	if (0 == strcasecmp(str,"magenta"))
		return magenta;
	if (0 == strcasecmp(str,"red"))
		return red;
	if (0 == strcasecmp(str,"white"))
		return white;
	if (0 == strcasecmp(str,"yellow"))
		return yellow;
	if (0 == strcasecmp(str,"none"))
		return noattr;
	if (0 == strcasecmp(str,"default"))
		return noattr;
	return invalid_color;
}


void init_terminal()
{
	if (0 == isatty(STDOUT_FILENO)) {
		dbug("stdout is not a terminal\n");
		TtyMode = tty_none;
	}
	
	if (TtyMode == tty_none) {
		dbug("tty mode set to none - disabling color&status support\n");
		init_tty_none();
		return;
	}
	if (TtyMode == tty_ansi) {
		dbug("setting up for ansi terminal\n");
		init_tty_ansi();
	}
	if (TtyMode == tty_auto) {
#ifdef HAVE_TERMINFO
		char *ts;
		int err = 0;
		if (ERR == setupterm(0,STDOUT_FILENO,&err))
			return;
		switch (err) {
		case 1:
			dbug("terminal setup ok\n");
			break;
		case 0:
			dbug("no terminal - disabling color\n");
			break;
		case -1:
		default:
			warn("unable to load terminal database - disabling color\n");
		}
		ts = tigetstr("sc");
		if ((ts != 0) && (ts != (char *)-1))
			TiSave = ts;
		ts = tigetstr("rc");
		if ((ts != 0) && (ts != (char *)-1))
			TiRestore = ts;
		ts = tigetstr("ll");
		if ((ts != 0) && (ts != (char *)-1))
			TiCursorLL = ts;
		ts = tigetstr("el");
		if ((ts != 0) && (ts != (char *)-1))
			TiClrEol = ts;
		ts = tigetstr("sgr0");
		if ((ts != 0) && (ts != (char *)-1))
			TiNoAttr = ts;
		char *setaf = tigetstr("setaf");
		char *setf = tigetstr("setf");
		if (setaf && (setaf != (const char *)-1)) {
			/* ANSI terminal capabilities */
			dbug("initializing terminal with ANSI capabilities\n");
			TiBlack = Strdup(tparm(setaf,0/*BLACK*/));
			TiRed = Strdup(tparm(setaf,1/*RED*/));
			TiGreen = Strdup(tparm(setaf,2/*GREEN*/));
			TiYellow = Strdup(tparm(setaf,3/*YELLOW*/));
			TiBlue = Strdup(tparm(setaf,4/*BLUE*/));
			TiMagenta = Strdup(tparm(setaf,5/*MAGENTA*/));
			TiCyan = Strdup(tparm(setaf,6/*CYAN*/));
			TiWhite = Strdup(tparm(setaf,7/*WHITE*/));
		} else if (setf && (setf != (const char *)-1)) {
			/* NON-ANSI terminal capabilities */
			dbug("initializing terminal with non-ANSI capabilities\n");
			TiBlack = Strdup(tparm(setaf,0/*BLACK*/));
			TiBlue = Strdup(tparm(setf,1/*BLUE*/));
			TiGreen = Strdup(tparm(setf,2/*GREEN*/));
			TiCyan = Strdup(tparm(setf,3/*CYAN*/));
			TiRed = Strdup(tparm(setf,4/*RED*/));
			TiMagenta = Strdup(tparm(setf,5/*MAGENTA*/));
			TiYellow = Strdup(tparm(setf,6/*YELLOW*/));
			TiWhite = Strdup(tparm(setf,7/*WHITE*/));
		}
#else
		dbug("terminfo is missing - disabling color&status support\n");
		init_tty_none();
#endif
	}
	size_t cll = strlen(TiClrEol);
	
	PfxWarn = makestr("%s%swarning:",TiClrEol,colorstr(ColorWarn));
	PfxStart = concatstr(TiClrEol,colorstr(ColorStart));
	PfxOut = concatstr(TiClrEol,colorstr(ColorOut));
	PfxError = concatstr(TiClrEol,colorstr(ColorError));
	PfxFail = concatstr(TiClrEol,colorstr(ColorFail));
	PfxDone = concatstr(TiClrEol,colorstr(ColorDone));
}


