/*
 * (c) 2006-2017 Thomas Maier-Komor
 * LICENSE: GPLv2, see file COPYING for details
 */

#include "config.h"

#include <assert.h>
#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#elif defined HAVE_STDLIB_H
# include <stdlib.h>
#endif
#ifdef HAVE_ALLOCA_H
# include <alloca.h>
#elif !defined alloca
# ifdef __GNUC__
#  define alloca __builtin_alloca
# elif defined _AIX
#  define alloca __alloca
# elif !defined HAVE_ALLOCA
#  ifdef  __cplusplus
extern "C"
#  endif
void *alloca (size_t);
# endif
#endif

#ifdef HAVE_MMAP
#include <sys/mman.h>
#endif

#ifdef HAVE_SPAWN
#include <spawn.h>
#endif

#ifdef HAVE_VFORK_H
#include <vfork.h>
#endif

#ifdef HAVE_WAIT4
#include <sys/resource.h>
#endif

#include "colortty.h"
#include "jobctrl.h"
#include "log.h"
#include "settings.h"
#include "support.h"

extern volatile int Terminate;
extern int ExitCode;

job_t *Executing = 0, *Ready = 0, *ReadyEnd = 0, *JobCache = 0;
arg_t *ArgCache = 0;
unsigned Skipped = 0, Finished = 0, Numbaseargs = 0, Running = 0,
	 Waiting = 0, Failed = 0, Parsed = 0;

static arg_t *Baseargs = 0;


void add_basearg(arg_t *a)
{
	a->prev = Baseargs;
	Baseargs = a;
	++Numbaseargs;
}


void name_job(job_t *j)
{
	char **a = j->args, *c;
	size_t l = 0;
	if (j->cmd)
		return;
	assert(a);
	do {
		l += strlen(*a);
		++l;
	} while (*++a);
	j->cmd = c = Malloc(l);
	a = j->args;
	do {
		c = (char *) memccpy(c,*a,0,l);
		*--c = ' ';
		++c;
	} while (*++a);
	--c;
	*c = 0;
}


void add_job(arg_t *a, size_t n, char *in, char *out, char *err, char *pwd, int app)
{
	char **argv, *exe;
	int i = n + Numbaseargs + 1;
	job_t *j;

	dbug("add_job(%p,%d,\"%s\",\"%s\",\"%s\",\"%s\",%d)\n",a,n,in?in:"",out?out:"",err?err:"",pwd?pwd:"",app);
	assert(a || n == 0);
	j = JobCache;
	if (j)
		JobCache = j->next;
	else
		j = Malloc(sizeof(job_t));
	argv = Malloc(sizeof(char *) * i);
	j->in = in;
	j->err = err;
	j->out = out;
	j->pwd = pwd;
	j->flags = app;
	j->outfile = -1;
	j->args = argv;
	j->pid = 0;
	j->next = 0;
	j->cmd = 0;
	--i;
	argv[i] = 0;
	j->numargs = i;
	while (n) {
		arg_t *c = a;
		assert(a);
		a = c->prev;
		--n;
		--i;
		argv[i] = c->arg;
		c->prev = ArgCache;
		ArgCache = c;
	}
	assert(a == 0);
	a = Baseargs;
	while (i) {
		arg_t *c = a;
		assert(a);
		a = c->prev;
		--i;
		argv[i] = c->arg;
	}
	assert(a == 0);
	exe = complete_exe(argv[0]);
	if (0 == exe) {
		warn("cannot find executable %s\n",argv[0]);
		j->pid = -1;
		exe = argv[0];
	}
	if ((argv[0] != exe) && (Numbaseargs == 0))
		free(argv[0]);
	argv[0] = exe;
	if (Verbose >= Debug) {
		if (in)
			dbug("stdin: %s\n",in);
		if (out)
			dbug("stdout: %s\n",out);
		if (err)
			dbug("stderr: %s\n",err);
		while (*argv) {
			dbug("argument %ld: <%s>\n",(long)(argv - j->args), *argv);
			++argv;
		}
	}
	j->next = 0;
	if (ReadyEnd)
		ReadyEnd->next = j;
	else
		Ready = j;
	ReadyEnd = j;
	dbug("%d jobs still waiting in the queue\n",++Waiting);
	++Parsed;
	start_jobs();
	if (Ready)
		dbug("job queued\n");
}


static void display_status()
{
	if (TtyMode != tty_none)
		status("*** %s%u queued, %s%u finished, %s%u failed, %s%u running%s\r"
			, TiYellow
			, Waiting
			, colorstr(ColorDone)
			, Finished
			, colorstr(ColorFail)
			, Failed
			, TiGreen
			, Running
			, TiNoAttr);
}


static void display_summary(job_t *j, int ret, long long dt, struct rusage *rusage, int showcommand)
{
	char real[64],user[64],sys[64];
	char *s;
	int fail;

#if defined(HAVE_ALLOCA_H)
	char *msg;
	msg = alloca(strlen(j->cmd) + 128);
#else
	char msg[1024];
#endif
	s = msg;
	if (ret)
		s = stpcpy(msg,PfxFail);
	else
		s = stpcpy(msg,PfxDone);
	if (showcommand)
		s += sprintf(s,"### job #%d (%s), pid %6ld, ",j->job,j->cmd,(long)j->pid);
	else
		s += sprintf(s,"### ");
	if (WIFSIGNALED(ret))
		s += sprintf(s,"terminated by signal %d (%s), ",WTERMSIG(ret),strsignal(WTERMSIG(ret)));
	else
		s += sprintf(s,"exitcode %d, ",WEXITSTATUS(ret));
	s += sprintf(s,"real: %s",timestr(dt,real));
	if (rusage && RsrcUsage) {
		if (rusage->ru_utime.tv_sec || rusage->ru_utime.tv_usec > 1000)
			s += sprintf(s,", user: %s",timevstr(rusage->ru_utime,user));
		if (rusage->ru_stime.tv_sec || rusage->ru_stime.tv_usec > 1000)
			s += sprintf(s,", sys: %s",timevstr(rusage->ru_stime,sys));
	}
	s = stpcpy(s,TiBlack);
	if (WIFSIGNALED(ret) || WEXITSTATUS(ret)) 
		status("%s%s\n",PfxFail,msg);
	else
		status("%s%s\n",PfxDone,msg);
}


void clear_job(pid_t pid, int ret, struct rusage *rusage)
{
	job_t *j = Executing, *lj = 0;
	char **args;
	struct stat st;
	unsigned long long now;
	long long dt;

	dbug("clear_job(%d,%d)\n",pid,ret);
	while (j && (j->pid != pid)) {
		lj = j;
		j = j->next;
	}
	if (j == 0) {
		/* consider the following bourne shell-script:
		 * #!/bin/sh
		 * ls -1 | ./a.out
		 *********************
		 * In this case a.out might be exec'ed by the shell
		 * _after_ forking childs for the other parts of the
		 * pipe. In consequence, depending on implementation,
		 * a.out can get a return value != -1 from wait that is
		 * associated with a process that cannot be found in our
		 * job list.  Be aware that the above script explicitly
		 * refers to /bin/sh not to /bin/bash. I.e. this issue
		 * concerns the original bourne shell.
		 */
		return;
	}
	now = gettime();
	dt = now-j->start;
	if (lj)
		lj->next = j->next;
	else
		Executing = j->next;
	st.st_size = 0;
	if (j->outfile != -1) {
		if (-1 == fstat(j->outfile,&st))
			warn("could not stat output file: %s\n",strerror(errno));
	}
#if HAVE_MMAP
	if (st.st_size > Pagesize) {
		void *mem = mmap(0,st.st_size,PROT_READ,MAP_SHARED,j->outfile,0);
		if (mem != (void *)-1) {
			int r;
			info("%s### output of job #%d, pid %ld: %s\n%s",ret?PfxFail:PfxDone,j->job,pid,j->cmd,PfxOut);
			r = fflush(stdout);
			if (r != 0)
				warn("unable to flush stdout: %s\n",strerror(errno));
			r = write(STDOUT_FILENO,mem,st.st_size);
			if (-1 == r)
				warn("unable to write output of job %d to stdout: %s\n",j->job,strerror(errno));
			r = munmap(mem,st.st_size);
			assert(r == 0);
			display_summary(j,ret,dt,rusage,0);
		} else {
			warn("could not mmap output of child: %s\n",strerror(errno));
			st.st_size = 0;
			display_summary(j,ret,dt,rusage,0);
		}
	} else
#endif
	if (st.st_size > 0) {
		int r, s0 = 0;
		size_t bsize = st.st_size + strlen(j->cmd) + 200;
#if __STDC_VERSION__ >= 199901L
		char buf[bsize];
#elif defined(HAVE_ALLOCA_H)
		char *buf;
		buf = alloca(bsize);
#else
		char *buf,*a;
		a = buf = Malloc(bsize);
#endif
		if (Verbose >= Info) {
			s0 = snprintf(buf,bsize,"%s### output of job #%d, pid %ld: %s\n%s",ret?PfxFail:PfxDone,j->job,(long)pid,j->cmd,PfxOut);
		}
		r = pread(j->outfile,buf+s0,st.st_size,0);
		if (r == -1)
			warn("error reading intermediate output file: %s\n",strerror(errno));
		else
			s0 += r;
		r = write(STDOUT_FILENO,buf,s0);
		if (r == -1)
			warn("unable to write result of job %d to stdout: %s\n",j->job,strerror(errno));
		display_summary(j,ret,dt,rusage,0);
#if (__STDC_VERSION__ < 199901L) && !defined(HAVE_ALLOCA_H)
		free(a);
#endif
	} else {
		display_summary(j,ret,dt,rusage,1);
	}
	if (ret != 0) {
		++Failed;
		if (ExitOnError) {
			warn("%s### job failed - terminating...\n",PfxFail);
			Terminate = 1;
		}
	}
	if (j->outfile != -1)
		(void) close(j->outfile);
	args = &j->args[Numbaseargs];
	while (*args) {
		free(*args);
		++args;
	}
	++Finished;
	free(j->cmd);
	free(j->args);
	assert(j != JobCache);
	j->next = JobCache;
	JobCache = j;
	dbug("%d jobs running\n",--Running);
	display_status();
}


static int ask_exec(job_t *j)
{
	int r;
	char ans[128];
	if (j->pid == -1)
		return 1;
	if (0 == Prompt)
		return 0;
	name_job(j);
	r = write(STDERR_FILENO,"start job '",11);
	if (-1 == r)
		warn("cannot write to terminal: %s\n",strerror(errno));
	r = write(STDERR_FILENO,j->cmd,strlen(j->cmd));
	if (-1 == r)
		warn("cannot write to terminal: %s\n",strerror(errno));
	r = write(STDERR_FILENO,"'? (hit y for yes, n for no)\n",29);
	if (-1 == r)
		warn("cannot write to terminal: %s\n",strerror(errno));
	do {
		r = read(STDERR_FILENO,ans,sizeof(ans));
		dbug("read stderr: %d : %s\n",r,ans);
		if ((-1 == r) && (errno == EINTR))
			continue;
		if (-1 == r) {
			warn("cannot read from terminal: %s\n",strerror(errno));
			j->pid = 0;	/* default to do not execute */
			return 1;
		}
		if ((0 == strncasecmp(ans,"n\n",2)) || (0 == strncasecmp(ans,"no\n",3))) {
			r = write(STDERR_FILENO,"no\n",3);
			if (-1 == r)
				warn("cannot write to terminal: %s\n",strerror(errno));
			j->pid = 0;
			return 1;
		}
	} while ((0 != strncasecmp(ans,"y\n",2)) && (0 != strncasecmp(ans,"yes\n",4)));
	r = write(STDERR_FILENO,"yes\n",4);
	if (-1 == r)
		warn("cannot write to terminal: %s\n",strerror(errno));
	return 0;
}


static void exec_job(job_t *j)
{
	static unsigned jobid = 0;
	int in = -1, out = -1, err = -1, log = -1;
	char pwd[PATH_MAX];
#ifdef HAVE_SPAWN
	posix_spawn_file_actions_t acts;

	ret = posix_spawn_file_actions_init(&acts);
	assert(ret == 0);
#endif
	if (ask_exec(j))
		return;
	if (j->pwd) {
		char *p = getcwd(pwd,sizeof(pwd));
		assert(p);
		if (-1 == chdir(j->pwd)) {
			warn("cannot change directory to %s: %s\n",j->pwd,strerror(errno));
			j->pid = -1;
			goto finish;
		}
	} else
		pwd[0] = 0;
	j->start = gettime();
	if (j->in) {	/* stdin is redirected */
		in = open(j->in,O_RDONLY,0666);
		if (in == -1) {
			warn("cannot open input file %s: %s\n",j->in,strerror(errno));
			j->pid = -1;
			return;
		}
		dbug("job input redirected to %s\n",j->in);
	} else {
		in = Stdin;
	}
	if (j->err) {	/* stdout and stderr are redirected */
		int mode;
		if (j->flags & 2)
			mode = O_WRONLY|O_CREAT|O_APPEND;
		else if (j->flags & 4)
			mode = O_WRONLY|O_CREAT|O_TRUNC;
		else
			mode = O_WRONLY|O_CREAT|O_TRUNC|O_EXCL;
		err = open(j->err,mode,0666);
		if (err == -1) {
			warn("cannot open error output file %s: %s\n",j->err,strerror(errno));
			j->pid = -1;
			goto finish;
		}
		out = err;
		dbug("job error output redirected to %s\n",j->err);
	} else if (Stdout != -1) {
		err = Stderr;
	}
	if (j->out) {	/* output is redirected */
		int mode;
		if (j->flags & 1)
			mode = O_WRONLY|O_CREAT|O_APPEND;
		else if (j->flags & 4)
			mode = O_WRONLY|O_CREAT|O_TRUNC;
		else
			mode = O_WRONLY|O_CREAT|O_TRUNC|O_EXCL;
		out = open(j->out,mode,0666);
		if (out == -1) {
			warn("cannot open output file %s: %s\n",j->out,strerror(errno));
			j->pid = -1;
			goto finish;
		}
		dbug("job output redirected to %s\n",j->out);
	} else if (Stdout != -1) {
		out = Stdout;
	}
	if (err == -1) {
		char fname[] = "/tmp/xjobs-XXXXXX";
		err = mkstemp(fname);
		if (-1 == err) {
			warn("cannot create intermediate output file %s: %s\n",fname,strerror(errno));
			j->pid = -1;
			goto finish;
		}
		dbug("stdout/stderr will be written to %s\n",fname);
		j->outfile = err;
		close_onexec(err);
		log = err;
		if (-1 == unlink(fname))
			warn("unlinking of temporary file %s failed: %s\n",fname,strerror(errno));
		if (out == -1)
			out = err;
	}

#ifdef HAVE_SPAWN
	ret = posix_spawn_file_actions_adddup2(&acts,err,STDERR_FILENO);
	assert(ret == 0);
	ret = posix_spawn_file_actions_adddup2(&acts,out,STDOUT_FILENO);
	assert(ret == 0);
	ret = posix_spawn_file_actions_adddup2(&acts,in,STDIN_FILENO);
	assert(ret == 0);
	if (posix_spawn(&j->pid,j->args[0],&acts,0,j->args,Env) == 0) {
		j->job = ++job;
		name_job(j);
		info("%s### started job #%d, pid %ld: %s\n",PfxStart,job,j->pid,j->cmd);
		j->next = Executing;
		Executing = j;
		dbug("started job, %d jobs waiting, %d running\n",--Waiting,++Running);
	} else {
		warn("couldn't spawn job %s: %s\n",j->cmd,strerror(errno));
		j->pid = -1;
	}
#elif defined(HAVE_WORKING_VFORK)
	j->pid = vfork();
	if (j->pid == -1) {
		warn("error could not vfork: %s\n",strerror(errno));
		goto finish;
	}
	if (j->pid != 0) {
		j->job = ++jobid;
		name_job(j);
		info("%s### started job #%d, pid %ld: %s\n",PfxStart,jobid,j->pid,j->cmd);
		j->next = Executing;
		Executing = j;
		dbug("started job, %d jobs waiting, %d running\n",--Waiting,++Running);
	} else {
		int r;
		char msg[256];
		if (in != -1)
			(void) dup2(in,STDIN_FILENO);
		if (out != -1)
			(void) dup2(out,STDOUT_FILENO);
		if (err != -1)
			(void) dup2(err,STDERR_FILENO);
		r = execve(j->args[0],j->args,Env);
		assert(r == -1);
		sprintf(msg,"failed to execute %s: %s\n",j->args[0],strerror(errno));
		r = write(STDERR_FILENO,msg,strlen(msg));
		assert(r != -1);
		_exit(EXIT_FAILURE);
	}
#else
	j->pid = fork();
	if (j->pid == -1) {
		warn("error could not fork: %s\n",strerror(errno));
		goto finish;
	}
	if (j->pid != 0) {
		j->job = ++job;
		name_job(j);
		info("%s### started job #%d, pid %ld: %s\n",PfxStart,job,j->pid,j->cmd);
		j->next = Executing;
		Executing = j;
		dbug("started job, %d jobs waiting, %d running\n",--Waiting,++Running);
	} else {
		int r;
		if (in != -1) {
			if (-1 == dup2(in,STDIN_FILENO))
				warn("unable to redirect stdin: %s\n",strerror(errno));
		}
		if (out != -1) {
			if (-1 == dup2(out,STDOUT_FILENO))
				warn("unable to redirect stdout: %s\n",strerror(errno));
		}
		if (err != -1) {
			if (-1 == dup2(err,STDERR_FILENO))
				warn("unable to redirect stderr: %s\n",strerror(errno));
		}
		r = execve(j->args[0],j->args,Env);
		assert(r == -1);
		warn("failed to execute %s: %s\n",j->args[0],strerror(errno));
		exit(EXIT_FAILURE);
	}
#endif

finish:
#ifdef HAVE_SPAWN
	(void) posix_spawn_file_actions_destroy(&acts);
#endif
	if ((in != -1) && (in != Stdin))
		(void) close(in);
	if ((out != -1) && (out != Stdout) && (out != log))
		(void) close(out);
	if ((err != -1) && (err != Stderr) && (err != out) && (err != log))
		(void) close(err);
	if (pwd[0]) {
		int r = chdir(pwd);
		if (r == -1)
			warn("unable to change back to working directory %s: %s\n",pwd,strerror(errno));
	}
}


int start_jobs(void)
{
	int num = 0;
	while ((Running < Limit) && Ready) {
		job_t *j = Ready;
		if ((j->args == 0) && (Running > 0)) {
			/* sequence point */
			dbug("sequence point - waiting for all processes to finish\n");
			return 0;
		}
		Ready = j->next;
		if (Ready == 0)
			ReadyEnd = 0;
		if (j->args) {
			exec_job(j);
			free(j->out);
			free(j->err);
			free(j->in);
			if (j->pid == 0) {
				warn("### skipping job: %s\n", j->cmd ?  j->cmd : j->args[0]);
				++Skipped;
			} else if (j->pid == -1) {
				char **args;
				warn("### error starting job: %s\n", j->cmd ?  j->cmd : j->args[0]);
				ExitCode = EXIT_FAILURE;
				args = &j->args[Numbaseargs];
				while (*args) {
					free(*args);
					++args;
				}
				free(j->cmd);
				free(j->args);
				assert(j != JobCache);
				j->next = JobCache;
				JobCache = j;
			}
			display_status();
		} else {
			dbug("starting jobs after sequence point\n");
			assert(j != JobCache);
			j->next = JobCache;
			JobCache = j;
		}
		++num;
	}
	display_status();
	return num;
}


