/*
 * (c) 2006-2017 Thomas Maier-Komor
 * LICENSE: GPLv2, see file COPYING for details
 */

#include <errno.h>
#include <limits.h>
#include <signal.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "support.h"
#include "tokens.h"
#include "log.h"


void *Malloc(size_t s)
{
	int tries = 0, time = 5;
	for (;;) {
		void *r = malloc(s);
		if (r != 0)
			return r;
		warn("out of memory - retrying in %d seconds\n",time);
		(void) sleep(time);
		time += 5;
		++tries;
		if (tries == 10) {
			warn("Persisiting low memory situation: stopping.\n"
				"Please send SIGCONT to contiue.\n");
			if (-1 == raise(SIGSTOP))
				abort();
			time = 5;
			tries = 0;
		}
	}
}


char *makestr(const char *f, ...)
{
	va_list val;
	va_start(val,f);
	size_t l = vsnprintf(0,0,f,val);
	va_end(val);
	++l;
	char *r = Malloc(l);
	va_start(val,f);
	vsnprintf(r,l,f,val);
	va_end(val);
	return r;
}


char *concatstr(const char *s0, const char *s1)
{
	size_t l0 = strlen(s0), l1 = strlen(s1);
	char *ret = Malloc(l0 + l1 + 1);
	memcpy(ret,s0,l0);
	memcpy(ret+l0,s1,l1+1);
	return ret;
}


static char get_stdin_char(void)
{
	static char buf[8*1024];
	static char *at = buf, *end = buf;
	if (at == end) {
		int num = read(STDIN_FILENO,buf,sizeof(buf));
		if (-1 == num) {
			fprintf(stderr,"error reading stdin: %s\n",strerror(errno));
			return 0;
		} else if (0 == num) {
			return 0;
		} else {
			at = buf;
			end = buf+num;
		}
	}
	return *at++;
}


int read_to_0(void)
{
	static char buf[PATH_MAX+1] = "";
	char *b = buf, c;
	if (buf[0] != 0) {
		buf[0] = 0;
		return EOL;
	}
	do {
		c = get_stdin_char();
		*b = c;
		++b;
		assert(b-buf < sizeof(buf));
	} while (c);
	if (*buf == 0)
		return 0;
	Fill = b - buf;
	Buffer = buf;
	return UQUOTED;
}


int read_to_nl(void)
{
	static char buf[PATH_MAX+1] = "";
	char *b = buf, c;
	if (buf[0] != 0) {
		buf[0] = 0;
		return EOL;
	}
	do {
		c = get_stdin_char();
		*b = c;
		++b;
		assert(b-buf < sizeof(buf));
	} while (c && (c != '\n'));
	if (*buf == 0)
		return 0;
	*--b = 0;
	Fill = b - buf;
	Buffer = buf;
	return UQUOTED;
}


int resolve_symlink(char *cmd)
{
	char lt[PATH_MAX+1];
	ssize_t l;
	/* handle symlinks */
	bzero(lt,sizeof(lt));
	while (-1 != (l = readlink(cmd,lt,PATH_MAX))) {
		char *base = cmd;
		lt[l] = '\0';
		dbug("is symlink: %s\n",lt);
		/* handle relative links */
		if (lt[0] != '/') {
			base = strrchr(cmd,'/');   
			if (base == 0) {
				base = cmd;
			} else if ((l + (base - cmd)) > PATH_MAX) {
				warn("command exceed PATH_MAX limit of your system\n");
				return -1;
			}
			++base;
		}
		(void) memcpy(base,lt,l+1);
		dbug("resolved to: %s\n",cmd);
	}
	return 0;
}


char *complete_exe(char *exe)
{
	typedef struct cache_s {
		struct cache_s *next;
		char *file, exe[PATH_MAX];
	} cache_t;
	static cache_t *cache = 0;
	cache_t *c = cache;
	char *token, *next;
	size_t exel;
	static char *Path = 0;

	if (Path == 0) {
		Path = getenv("PATH");
		dbug("PATH=\"%s\"\n", Path);
	}
	while (c) {
		if (0 == strcmp(exe,c->exe)) {
			if (c->file == c->exe)
				return exe;
			return strdup(c->file);
		}
		c = c->next;
	}
	if ((exe[0] == '/') && (0 == access(exe,X_OK))) {
		c = Malloc(sizeof(cache_t));
		strcpy(c->exe,exe);
		c->file = c->exe;
		c->next = cache;
		cache = c;
                return exe;
	}
	exel = strlen(exe);
	if ((strncmp(exe,"./",2) == 0) || (strncmp(exe,"../",3) == 0)) {
		if (0 == access(exe,X_OK)) {
			c = Malloc(sizeof(cache_t));
			strcpy(c->exe,exe);
			c->file = c->exe;
			c->next = cache;
			cache = c;
			return exe;
		}
		if (Verbose == Debug) {
			char cmd[PATH_MAX+1];
			assert(exel < sizeof(cmd));
			(void) memcpy(cmd,exe,exel);
			(void) resolve_symlink(cmd);
		}
		return 0;
	}
	token = Path;
	next = strchr(Path,':');
        while (token && *token) {
		char cmd[PATH_MAX];
		size_t off;

		if (next)
			*next = 0;
		if ((off = strlen(token)) >= PATH_MAX) {
			warn("command exceed PATH_MAX limit of your system\n");
			return 0;
		}
		(void) memcpy(cmd, token, off+1);
		if (cmd[off-1] != '/') {
			cmd[off] = '/';
			++off;
		}
		if (exel + off >= PATH_MAX) {
			warn("command exceed PATH_MAX limit of your system\n");
			return 0;
		}
		(void) memcpy(cmd+off, exe, exel+1);
		if (next) {
			*next = ':';
			token = next + 1;
			next = strchr(token,':');
		} else
			token = 0;
		dbug("candidate: %s\n",cmd);
		if ((Verbose == Debug) && resolve_symlink(cmd))
			continue;
		if (0 == access(cmd,X_OK)) {
			dbug("using: %s\n",cmd);
			c = Malloc(sizeof(cache_t));
			strcpy(c->exe,exe);
			c->file = (char *) strdup(cmd);
			c->next = cache;
			cache = c;
                        return strdup(cmd);
		}
        }
        return 0;
}


const char *timestr(long long usec, char *ustr)
{
	static char str[64];
	char *s = ustr ? ustr : str;
	unsigned days, hours, mins, secs;
	double sec;
	secs = usec / 1000000L;
	usec = usec % 1000000L;
	days = secs / (60 * 60 * 24);
	secs -= days * 60 * 60 * 24;
	hours = secs / (60 * 60);
	secs -= hours * 60 * 60;
	mins = secs / 60;
	secs -= mins * 60;
	assert(secs < 60);
	sec = (double)secs + (double)usec * 1E-6;
	if (days)
		s += sprintf(s,"%dd %dh %dm",days,hours,mins);
	else if (hours)
		s += sprintf(s,"%dh %dm",hours,mins);
	else if (mins)
		s += sprintf(s,"%dm %ds",mins,secs);
	else
		s += sprintf(s,"%.4gs",sec);
	return ustr ? ustr : str;
}


