/*
 * (c) 2006-2017 Thomas Maier-Komor
 * LICENSE: GPLv2, see file COPYING for details
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <math.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#ifdef HAVE_STROPTS_H
#include <stropts.h>
#endif

#ifdef HAVE_WAIT4
#include <sys/resource.h>
#endif


/* SUSv3 does not have PATH_MAX */
#ifndef PATH_MAX
#define PATH_MAX _XOPEN_PATH_MAX
#endif

#if __STDC_VERSION__ < 199901L
#define inline
#endif

#ifndef SIGPOLL
#define SIGPOLL SIGIO
#endif

#include "log.h"
#include "tokens.h"
#include "support.h"
#include "jobctrl.h"
#include "settings.h"
#include "colortty.h"


extern int yylex(void);

char **Env;
int ExitCode = EXIT_SUCCESS, InputsWaiting = 0, Terminate = 0;
volatile int ChildTerminated = 0;

static unsigned long long Start = 0;
static pid_t PGID = 0;


RETSIGTYPE processSignal(int sig)
{
	switch (sig) {
	case SIGPOLL:
		dbug("SIGPOLL\n");
		++InputsWaiting;
		break;
	case SIGTERM:
		dbug("SIGTERM\n");
		Terminate = 2;
		ExitCode = EXIT_FAILURE;
		break;
	case SIGINT:
		dbug("SIGINT\n");
		if (++Terminate == 2) {
			if (TiNoAttr && TiNoAttr[0])
				write(STDOUT_FILENO,TiNoAttr,strlen(TiNoAttr));
			_exit(EXIT_FAILURE);
		}
		break;
	case SIGCHLD:
		dbug("SIGCHLD\n");
		++ChildTerminated;
		break;
	default:
		abort();
	}
}


void finalize(void)
{
	while (Executing) {
		pid_t pid;
		int ret;
#ifdef HAVE_WAIT4
		struct rusage rusage;
		pid = wait4(0,&ret,0,&rusage);
		if ((pid == -1) && (errno == EINTR))
			continue;
		clear_job(pid,ret,&rusage);
#else
		pid = wait(&ret);
		if ((pid == -1) && (errno == EINTR))
			continue;
		clear_job(pid,ret,0);
#endif
		if (pid == -1) {
			warn("error waiting for children: %s\n",strerror(errno));
			ExitCode = EXIT_FAILURE;
			break;
		}
	}
	info("summary: %s%u failed, %s%u succeeded, %sexecuted in %s\n",PfxFail,Failed,PfxDone,Finished-Failed,TiNoAttr,timestr(gettime()-Start,0));
	exit(ExitCode);
}


void parse_args(int argc, char **argv)
{
	while (optind < argc) {
		if (0 == strcmp("<",argv[optind])) {
			if (Stdin != -1)
				(void) close(Stdin);
			if (argv[++optind] == 0)
				error("missing argument for input redirector <\n");
			Stdin = open(argv[++optind],O_RDONLY);
			if (-1 == Stdin)
				error("could not open default stdin for jobs (%s): %s\n",argv[optind],strerror(errno));
		} else if (0 == strcmp(">",argv[optind])) {
			if (Stdout != -1)
				(void) close(Stdout);
			if (argv[++optind] == 0)
				error("missing argument for output redirector >\n");
			Stdout = open(argv[optind],O_WRONLY|O_CREAT|O_EXCL,0666);
			if (-1 == Stdout)
				error("could not open stdout for jobs (%s): %s\n",argv[optind],strerror(errno));
		} else if (0 == strcmp(">>",argv[optind])) {
			if (Stdout != -1)
				(void) close(Stdout);
			if (argv[++optind] == 0)
				error("missing argument for output redirector >>\n");
			Stdout = open(argv[optind],O_WRONLY|O_APPEND);
			if (-1 == Stdout)
				error("could not append to stdout for jobs (%s): %s\n",argv[optind],strerror(errno));
		} else if (0 == strcmp(">&",argv[optind])) {
			if (Stdout != -1)
				(void) close(Stdout);
			if (argv[++optind] == 0)
				error("missing argument for output redirector >&\n");
			Stderr = Stdout = open(argv[optind],O_WRONLY|O_CREAT|O_EXCL,0666);
			if (-1 == Stdout)
				error("could not open stdout/stderr for jobs (%s): %s\n",argv[optind],strerror(errno));
		} else if (0 == strcmp(">>&",argv[optind])) {
			if (Stdout != -1)
				(void) close(Stdout);
			if (argv[++optind] == 0)
				error("missing argument for output redirector >>&\n");
			Stdout = open(argv[optind],O_WRONLY|O_APPEND);
			if (-1 == Stdout)
				error("could not append to stdout/stderr for jobs (%s): %s\n",argv[optind],strerror(errno));
		} else {
			arg_t *a = Malloc(sizeof(arg_t));
			if (Numbaseargs == 0) {
				a->arg = complete_exe(argv[optind]);
				if (a->arg == 0)
					error("cannot find %s in PATH\n",argv[optind]);
			} else {
				a->arg = argv[optind];
			}
			add_basearg(a);
		}
		++optind;
	}
}


int main(int argc, char **argv, char **env)
{
	typedef enum { init, std, white, dquoted, squoted, desc, sesc } state_t;
	int i, flags = 0, ret;
	unsigned long lcount = 0;
	char *jobin = 0, *jobout = 0, *joberr = 0, *jobpwd = 0;
	arg_t *args = 0;
	size_t numargs = 0;
	struct sigaction sig;
	token_t t;

	Start = gettime();
	for (i = 1; i < argc; ++i) {
		const char *arg = argv[i];
		if ((arg[0] != '-') || (arg[1] != 'v'))
			continue;
		const char *verbose = 0;
		if (arg[2]) {
			verbose = arg+2;
		} else if (argv[++i][0] == '-') {
			warn("missing argument to option -v");
			continue;
		} else {
			verbose = argv[i];
		}
		long v = strtol(verbose,0,0);
		if (v < Silent || v > Debug)
			error("value %ld out of range for option -v\n",v);
		else
			Verbose = v;
	}
	Env = env;
	setenv("POSIXLY_CORRECT","1",1);
	init_limit();
	init_defaults(argv[0]);
	parse_options(argc,argv);
	parse_args(argc,argv);
	init_terminal();

	if (Stdin == -1) {	/* no default input redirector */
		Stdin = open("/dev/null",O_RDONLY);
		if (Stdin == -1)
			warn("could not open /dev/null: %s\n",strerror(errno));
	}

	close_onexec(Stdin);
	close_onexec(Stdout);
	close_onexec(Stderr);
	PGID = getpgid(getpid());

	sig.sa_handler = processSignal;
	sigemptyset(&sig.sa_mask);
	sigaddset(&sig.sa_mask,SIGTERM);
	sig.sa_flags = 0;
	ret = sigaction(SIGTERM,&sig,0);
	assert(ret == 0);
	sigemptyset(&sig.sa_mask);
	sigaddset(&sig.sa_mask,SIGINT);
	ret = sigaction(SIGINT,&sig,0);
	assert(ret == 0);
	sigemptyset(&sig.sa_mask);
	sigaddset(&sig.sa_mask,SIGCHLD);
	ret = sigaction(SIGCHLD,&sig,0);
	assert(ret == 0);
	while ((Terminate == 0) && (0 != (t = gettoken()) || Script)) {
		switch (t) {
		default:
			abort();
			break;
		case CD:
			jobpwd = Strdupn(Buffer,Fill+1);
			dbug("CD %s\n",jobpwd);
			continue;
		case SEQUPT:
			{
				job_t *j = Malloc(sizeof(job_t));
				j->args = 0;
				if (ReadyEnd)
					ReadyEnd->next = j;
				else
					Ready = j;
				ReadyEnd = j;
				continue;
			}
		case INPUT:
			t = yylex();
			if ((t != UQUOTED) && (t != DQUOTED) && (t != SQUOTED)) {
				warn("invalid argument for input redirector <\n");
				do t = yylex(); while (t && (t != EOL));
			} else {
				jobin = Strdupn(Buffer,Fill+1);
			}
			continue;
		case APPEND:
			flags |= 1;
			/*FALLTHROUGH*/
		case OUTPUT:
			t = yylex();
			if ((t != UQUOTED) && (t != DQUOTED) && (t != SQUOTED)) {
				warn("invalid argument for output redirector >\n");
				do t = yylex(); while (t && (t != EOL));
			} else {
				jobout = Strdupn(Buffer,Fill+1);
			}
			continue;
		case OVWROUTPUT:
			flags |= 4;
			t = yylex();
			if ((t != UQUOTED) && (t != DQUOTED) && (t != SQUOTED)) {
				warn("invalid argument for output redirector >!\n");
				do t = yylex(); while (t && (t != EOL));
			} else {
				jobout = Strdupn(Buffer,Fill+1);
			}
			continue;
		case APPERR:
			flags |= 3;
			/*FALLTHROUGH*/
		case ERRLOG:
			t = yylex();
			if ((t != UQUOTED) && (t != DQUOTED) && (t != SQUOTED)) {
				warn("invalid argument for log redirector >&\n");
				do t = yylex(); while (t && (t != EOL));
			} else {
				joberr = Strdupn(Buffer,Fill+1);
			}
			continue;
		case UQUOTED:
		case DQUOTED:
		case SQUOTED:
			{
				arg_t *a = ArgCache;
				if (a)
					ArgCache = a->prev;
				else
					a = Malloc(sizeof(arg_t));
				a->arg = Strdupn(Buffer,Fill+1);
				a->prev = args;
				args = a;
				++numargs;
				continue;
			}
		case EOL:
			if (++lcount < Lines)
				break;
			if (numargs || jobin || jobout || joberr)
				add_job(args,numargs,jobin,jobout,joberr,jobpwd,flags);
			args = 0;
			numargs = 0;
			jobin = 0;
			jobout = 0;
			joberr = 0;
			jobpwd = 0;
			flags = 0;
			lcount = 0;
			if (QLen && (QLen == Waiting))
				break;
			continue;
		case LEXERR:
			info("ignoring erroneous line\n");
			continue;
		case 0:
			dbug("token: 0\n");
			break;
		}

		if (start_jobs())
			continue;
		if (ChildTerminated) {
			int flags = WNOHANG, ret;
			struct sigaction orig;
			siginfo_t si;
			pid_t pid;
			--ChildTerminated;
			if (QLen && (QLen == Waiting)) {
				dbug("queue is full\n");
			} else {
				flags = WNOHANG;
			}
			if (InputsWaiting == 0) {
				dbug("wait blocking - listening for new pipe connection\n");
				sig.sa_handler = processSignal;
				sigemptyset(&sig.sa_mask);
				sigaddset(&sig.sa_mask,SIGPOLL);
				sig.sa_flags = 0;
				ret = sigaction(SIGPOLL,&sig,&orig);
				assert(ret == 0);
			}
#ifdef HAVE_WAIT4
			if (RsrcUsage) {
				struct rusage rusage;
				if (0 < (pid = wait4(0,&ret,InputsWaiting == 0 ? 0 : flags,&rusage))) 
					clear_job(pid,ret,&rusage);
			} else
#endif
#ifdef __sun
				if ((0 == waitid(P_PGID,PGID, &si, ((QLen && QLen == Waiting) ? 0 : WNOHANG) | WEXITED)) && (si.si_pid != 0))
					clear_job(si.si_pid,si.si_status,0);
#else

			if ((pid = waitpid(0,&ret,flags)) > 0) {
				clear_job(pid,ret,0);
			}
#endif
			if (InputsWaiting == 0) {
				ret = sigaction(SIGPOLL,&orig,0);
				assert(ret == 0);
			}
			start_jobs();
		}
	}
	if (numargs)
		add_job(args,numargs,jobin,jobout,joberr,jobpwd,flags);

	while ((Executing || Ready) && (Terminate == 0)) {
		pid_t pid;
		int ret;
#ifdef HAVE_WAIT4
		struct rusage rusage;
#endif
		start_jobs();
#ifdef HAVE_WAIT4
		pid = wait4(0,&ret,0,&rusage);
		clear_job(pid,ret,&rusage);
#else
		pid = wait(&ret);
		clear_job(pid,ret,0);
#endif
	}
	finalize();
}

/* vim:tw=0
 */
